
/*
 * Copyright (c) 2001-2003 Shiman Associates Inc. All Rights Reserved.
 * 
 * Permission is hereby granted, free of charge, to any person
 * obtaining a copy of this software and associated documentation
 * files (the "Software"), to deal in the Software without
 * restriction, including without limitation the rights to use, copy,
 * modify, merge, publish, distribute, sublicense, and/or sell copies
 * of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 * 
 * The above copyright notice and this permission notice shall be
 * included in all copies or substantial portions of the Software.
 * 
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 *
 */

#include <stdio.h>
#include <gtk/gtk.h>
#include <pango/pango.h>
#include "masdbscale.h"

/* $Id: masdbscale.c,v 1.2 2003/12/12 16:28:34 silvio Exp $ */

static GtkDrawingAreaClass *parent_class;


void gtk_masdbscale_set_db_cutoff(GtkWidget* w, int new_cutoff)
{
    GtkMasdbscale *scale = GTK_MASDBSCALE(w);
    scale->db_cutoff = new_cutoff;
}


static int
draw_stereo_notflush(GtkWidget *widget, GdkEventExpose *event)
{
    char strg[5];
    int i;
    int tw, th;
    int db_cutoff;
    int he, wi;
    int maxw = 0;
    GtkMasdbscale *scale = GTK_MASDBSCALE(widget);

    db_cutoff = scale->db_cutoff;    
    he = widget->allocation.height  - 2*MDS_GTKSLIDERAD - 1;
    wi = widget->allocation.width / 5;

    for ( i=0; i<=db_cutoff; i=scale->incr_func(i) )
    {
        sprintf(strg, "%d", scale->val_func(i));
        pango_layout_set_text( scale->layout, strg, -1 );
        pango_layout_get_pixel_size( scale->layout, &tw, &th );
        
        gdk_draw_line ( widget->window, widget->style->fg_gc[widget->state],
                        0, MDS_GTKSLIDERAD+(i*he)/db_cutoff, wi, MDS_GTKSLIDERAD+(i*he)/db_cutoff );
        gdk_draw_line ( widget->window, widget->style->fg_gc[widget->state],
                        widget->allocation.width-wi-1, MDS_GTKSLIDERAD+(i*he)/db_cutoff,
                        widget->allocation.width-1, MDS_GTKSLIDERAD+(i*he)/db_cutoff );

        sprintf(strg, "%d", i);

        pango_layout_set_text( scale->layout, strg, -1 );
        pango_layout_get_pixel_size( scale->layout, &tw, &th );

        gtk_paint_layout (widget->style,
                          widget->window,
                          GTK_WIDGET_STATE (widget),
                          FALSE,
                          &event->area,
                          widget,
                          "",
                          (widget->allocation.width-tw)/2,
                          MDS_GTKSLIDERAD+(i*he)/db_cutoff-th/2,
                          scale->layout);

        if( tw>maxw )
            maxw = tw;
    }
    
    return maxw;
}

static int
draw_stereo_flush(GtkWidget *widget, GdkEventExpose *event)
{
    char strg[5];
    int i;
    int tw, th;
    int db_cutoff;
    int he, wi, xpos, ypos;
    int maxw = 0;
    GtkMasdbscale *scale = GTK_MASDBSCALE(widget);

    db_cutoff = scale->db_cutoff;    
    he = widget->allocation.height - 1;
    wi = widget->allocation.width / 5;

    for ( i=0; i<=db_cutoff; i=scale->incr_func(i) )
    {
        sprintf(strg, "%d", scale->val_func(i));
        pango_layout_set_text( scale->layout, strg, -1 );
        pango_layout_get_pixel_size( scale->layout, &tw, &th );
        
        gdk_draw_line ( widget->window, widget->style->fg_gc[widget->state],
                        0, (i*he)/db_cutoff, wi, (i*he)/db_cutoff );
        gdk_draw_line ( widget->window, widget->style->fg_gc[widget->state],
                        widget->allocation.width-wi-1, (i*he)/db_cutoff,
                        widget->allocation.width-1, (i*he)/db_cutoff );

        sprintf(strg, "%d", i);

        pango_layout_set_text( scale->layout, strg, -1 );
        pango_layout_get_pixel_size( scale->layout, &tw, &th );

        xpos = (widget->allocation.width-tw)/2;
        ypos = (i*he)/db_cutoff-th/2;
        
        
        if( i==0 )
        {
            ypos = -3;
        }
        if( i==db_cutoff )
        {
            ypos = he - th + 3;
        }

        gtk_paint_layout (widget->style,
                          widget->window,
                          GTK_WIDGET_STATE (widget),
                          FALSE,
                          &event->area,
                          widget,
                          "",
                          xpos,
                          ypos,
                          scale->layout);

        if( tw>maxw )
            maxw = tw;
    }
    
    return maxw;
}


static int
draw_mono_notflush( GtkWidget *widget, GdkEventExpose *event )
{
    char strg[5];
    int i, widget_hs, widget_w3;
    int tw, th;
    int db_cutoff;
    int maxw = 0;
    GtkMasdbscale *scale = GTK_MASDBSCALE(widget);

    db_cutoff = scale->db_cutoff;

    widget_hs  = widget->allocation.height - 2*MDS_GTKSLIDERAD - 1;
    widget_w3 = widget->allocation.width / 3;
    
    for ( i=0; i<=db_cutoff; i=scale->incr_func(i) )
    {
        gdk_draw_line( widget->window, widget->style->fg_gc[widget->state],
                       widget->allocation.width-widget_w3-1, MDS_GTKSLIDERAD+(i*widget_hs)/db_cutoff,
                       widget->allocation.width-1, MDS_GTKSLIDERAD+(i*widget_hs)/db_cutoff );
        
        sprintf(strg, "%d", scale->val_func(i));
        pango_layout_set_text( scale->layout, strg, -1 );
        pango_layout_get_pixel_size( scale->layout, &tw, &th );

        if( tw>maxw )
            maxw = tw;
        
        gtk_paint_layout( widget->style,
                          widget->window,
                          GTK_WIDGET_STATE (widget),
                          FALSE,
                          &event->area,
                          widget,
                          "",
                          2*widget_w3-tw-2,
                          MDS_GTKSLIDERAD+(i*widget_hs)/db_cutoff-th/2,
                          scale->layout );
    }
    
    return maxw;
}


static int
draw_mono_flush( GtkWidget *widget, GdkEventExpose *event )
{
    char strg[5];
    int i, height, widget_w3;
    int tw, th;
    int db_cutoff, xpos, ypos;
    int maxw = 0;
    GtkMasdbscale *scale = GTK_MASDBSCALE(widget);

    db_cutoff = scale->db_cutoff;

    height  = widget->allocation.height-1;
    widget_w3 = widget->allocation.width / 3;
    
    
    for ( i=0; i<=db_cutoff; i=scale->incr_func(i) )
    {
        gdk_draw_line ( widget->window, widget->style->fg_gc[widget->state],
                        widget->allocation.width-widget_w3-1, (i*height)/db_cutoff,
                        widget->allocation.width-1, (i*height)/db_cutoff );
        
        sprintf(strg, "%d", scale->val_func(i));
        pango_layout_set_text( scale->layout, strg, -1 );
        pango_layout_get_pixel_size( scale->layout, &tw, &th );

        if( tw>maxw )
            maxw = tw;

        xpos = 2*widget_w3-tw-2;
        ypos = (i*height)/db_cutoff-th/2;
        
        if( i==0 )
        {
            ypos = -3;
        }
        if( i==db_cutoff )
        {
            ypos = height - th + 3;
        }

        gtk_paint_layout (widget->style,
                          widget->window,
                          GTK_WIDGET_STATE (widget),
                          FALSE,
                          &event->area,
                          widget,
                          "",
                          xpos,
                          ypos,
                          scale->layout);
    }
    
    return maxw;
}



static gboolean
draw_scale(GtkWidget *widget, GdkEventExpose *event)
{    
    GtkMasdbscale *scale = GTK_MASDBSCALE(widget);
    int maxw = 0;
    
    
    if( !scale->layout )
        scale->layout = gtk_widget_create_pango_layout( widget, "0");


    if( !scale->leave_colors )
    {
        gdk_window_clear_area (widget->window,
                               event->area.x, event->area.y,
                               event->area.width, event->area.height);
        
        gdk_gc_set_clip_rectangle( widget->style->fg_gc[widget->state],
                                   &event->area );
    }
    

    if( scale->isflush )
    {
        if( scale->ismono )
        {
            maxw = draw_mono_flush( widget, event );
        }
        else
        {
            maxw = draw_stereo_flush( widget, event );
        }
        
    }
    else
    {
        if( scale->ismono )
        {
            maxw = draw_mono_notflush( widget, event );
        }
        else
        {
            maxw = draw_stereo_notflush( widget, event );
        }
    }
    
    
        
    if( !scale->leave_colors )
    {
        gdk_gc_set_clip_rectangle (widget->style->fg_gc[widget->state], NULL);
    }

    if( maxw>scale->oldmaxw )
    {
        gtk_widget_set_size_request( widget, (3*maxw)/2+2, widget->allocation.height );
        scale->oldmaxw = maxw;
    }
    
    /*     gtk_widget_queue_resize( widget ); */

    return TRUE;
}


static gboolean
gtk_masdbscale_expose(GtkWidget *widget, GdkEventExpose *event)
{
    GtkMasdbscale *scale;
    scale = GTK_MASDBSCALE(widget);

    draw_scale( widget, event ) ;
    
    return TRUE;
}




static void
gtk_masdbscale_class_init (GtkMasdbscaleClass *klass)
{
    GtkWidgetClass *widget_class;
    widget_class = (GtkWidgetClass*) klass;
    
    widget_class->expose_event = gtk_masdbscale_expose;
}


static void
gtk_masdbscale_init( GtkMasdbscale *masdbscale )
{
}


guint
gtk_masdbscale_get_type()
{
  static guint mytype = 0;

  if (!mytype)
    {
      GtkTypeInfo lh_info =
      {
	"GtkMasdbscale",
	sizeof(GtkMasdbscale),
	sizeof(GtkMasdbscaleClass),
	(GtkClassInitFunc) gtk_masdbscale_class_init,
        (GtkObjectInitFunc) gtk_masdbscale_init,
        NULL,
        NULL
      };

      mytype = gtk_type_unique (gtk_drawing_area_get_type(), &lh_info);
    }

  return mytype;
}

GtkWidget*
gtk_masdbscale_new( struct mds_options *opts )
{
    GtkRcStyle *rc_style;
    GdkColor bgcolor, fgcolor;
    
    GtkMasdbscale *s = gtk_type_new (gtk_masdbscale_get_type ());

    s->incr_func = opts->incr_func;
    s->val_func = opts->val_func;
    s->ismono = opts->ismono;
    s->leave_colors = opts->leave_colors;
    s->db_cutoff = opts->db_cutoff;
    s->oldmaxw = 0;
    s->isflush = opts->isflush;
    
    if( !s->leave_colors )
    {
        bgcolor.red = 0;
        bgcolor.green = 0;
        bgcolor.blue = 0;
        
        fgcolor.red = ~0;
        fgcolor.green = ~0;
        fgcolor.blue = ~0;
        
        rc_style = gtk_rc_style_new ();
        rc_style->bg[GTK_STATE_NORMAL] = bgcolor;
        rc_style->fg[GTK_STATE_NORMAL] = fgcolor;
        rc_style->color_flags[GTK_STATE_NORMAL] |= GTK_RC_BG;
        rc_style->color_flags[GTK_STATE_NORMAL] |= GTK_RC_FG;
        gtk_widget_modify_style (GTK_WIDGET(s), rc_style);
        gtk_rc_style_unref (rc_style);
    }
    
    return GTK_WIDGET (s);
}
